/*
 *
 *  Copyright (C) 2015 Jürg Müller, CH-5524
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation version 3 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program. If not, see http://www.gnu.org/licenses/ .
 */

#if defined(__PYTHON__)
  #include <Python.h>
#endif

#include <stdio.h>
#include <string.h>

#if defined(__LINUX__)
	#include <utime.h>
	#include <unistd.h>   // für usleep
	#include <stdlib.h>
#endif

#include <sys/stat.h>
#if !defined(__VC__) && !defined(_MSC_VER)
  #include <dirent.h>
#endif

#include "NTypes.h"
#include "NUtils.h"

using namespace NUtils;

#include "VList.h"

bool VList::AddString(const char * string)
{
  if (!string)
    return false;

  if (FindString(string) >= 0)
    return false;

  unsigned long len = strlen(string) + 1;
  unsigned long old_size = mSize;
  unsigned long new_size = mSize;
  if (!string_count)
    new_size = len < 10000 ? 10000 : 2*len;
  else
    new_size += 2*len;
  SetSize(new_size);
  mSize = old_size;

  memcpy(mMemory + mSize, string, len);
  string_count++;
  mSize += len;
  mMemory[mSize] = 0;
  
  is_sorted = false;

  return true;
}

int VList::FindString(const char * string) const
{
  if (!mMemory || !string)
    return -1;

  unsigned index = 0;
  const char * res = (const char *) (mMemory);
  while (index < string_count)
  {
    if (!strcmp(res, string))
      return index;

    res += strlen(res) + 1;
    index++;
  }
  return -1;
}

bool VList::Sort()
{
  for (int i = 0; i < (int) string_count - 1; i++)
  {
    char * stri = (char *) GetString(i);
    const char * str = stri;
    int index = i;
    for (unsigned j = i+1; j < string_count; j++)
    {
      const char * s = GetString(j);
      if (stricmp(str, s) > 0)
      {
        str = s;
        index = j;
      }
    }
    if (index > i)
    {
      unsigned long old_size = mSize;
      long len = strlen(str) + 1;
      SetSize(old_size + len);
      mSize = old_size;
      long pos = stri - (const char *) mMemory;
      memmove(stri + len, stri, mSize - pos);
      str += len;
      memcpy(stri, str, len);
      pos = str - (const char *) mMemory;
      memmove((char *) str, str + len, mSize - pos);
    }
  }

  is_sorted = true;

  return true;
}

const char * VList::GetString(unsigned index)
{
  if (index >= string_count || !mMemory)
    return NULL;

  const char * res = (const char *) (mMemory);
  while (index)
  {
    res += strlen(res) + 1;
    index--;
  }
  return res;
}

bool VList::DeleteString(unsigned index)
{
  const char * pos = GetString(index);
  if (!pos)
    return false;
  
  long p = pos - (const char *) mMemory;
  long len = strlen(pos);
  
  memmove((char *) pos, pos+len+1, mSize-p+1);
  string_count--;
  mSize -= len+1;
  mMemory[mSize] = 0;
  
  return true;  
}

bool VList::ReadFile(const char * Filename, unsigned long Max)
{
  KStream stream;
  if (!stream.ReadFile(Filename, Max))
    return false;
  
  for (long p = 0; p < stream.Size(); p++)
    if (stream.GetByte(p) < ' ')
    {
      stream.SetByte(p, 0);
    }
  
  for (long p = 0; p < stream.Size(); )
  {
    const char * ptr = stream.GetCharMemory() + p;
    long l = strlen(ptr);
    if (l > 0)
      AddString(ptr);
    p += l + 1;
  }
  return true;
}

bool VList::SaveFile(const char * Filename) const
{
  KStream stream;
  
  stream.SetSize(Size() + 1);
  memcpy((char *) stream.GetMemory(), GetMemory(), Size());
  *((char *) stream.GetMemory() + Size()) = 0;
  
  for (long p = 0; p < stream.Size(); p++)
    if (stream.GetByte(p) == 0)
    {
      stream.SetByte(p, '\n');
    }
  
  return stream.SaveFile(Filename);  
}


////////////////////////////////////////////////////////////////////////////////

bool VExploreDir::Start(const char * StartFromDir)
{
  strcpy(FromPath, StartFromDir);
  MakeStandardPath(FromPath);

  RelPath[0] = 0;

  if (!NUtils::DirExists(FromPath))
  {
    printf("directory \"%s\" does not exist\n", FromPath);
    return false;
  }

  return ExploreDir("");
}

bool VExploreDir::ExploreSortedList(bool dirs)
{
  bool Ok = true;
  VListFileDir ListFile;
  VList sorted_list;

  if (ListFile.Open(FromPath, RelPath))
  {
    while (ListFile.Next())
    {
      if ((ListFile.Name[0] != '.' || (!dirs && use_hidden_files)) &&
          ListFile.Name[0] != '@')
      {
        if ((!dirs && ListFile.IsFile()) ||
            (dirs && ListFile.IsDir()))
        {
          if (!sorted_list.AddString(ListFile.Name))
            Ok = false;
        }
      }
    }
    ListFile.Close();
  }
  if (!sorted_list.Sort())
    Ok = false;

  if (Ok)
  {
    long pos = strlen(RelPath);
    for (int i = 0; i < (int) sorted_list.string_count; i++)
    {
      const char * name = sorted_list.GetString(i);
      if (!dirs)
      {
        if (!ExploreFile(name))
          Ok = false;
      } else
      if (pos + strlen(name) + 2 < sizeof(RelPath))
      {
        strcat(RelPath, name);
        strcat(RelPath, "/");
        if (!ExploreDir(name))
          Ok = false;
        RelPath[pos] = 0;
      }
    }
  }

  return Ok;
}

bool VExploreDir::ExploreDir(const char * DirName)
{
  bool Ok = ExploreSortedList(dir_first);
  if (!ExploreSortedList(!dir_first))
    Ok = false;

  return Ok;
}

VListFileDir::VListFileDir()
{
#if defined(__VC__) || defined(_MSC_VER)
  FindFileData = new WIN32_FIND_DATA;
  FirstRead = false;
#endif
  dir = NULL;
}

VListFileDir::~VListFileDir()
{
  Close();
#if defined(__VC__) || defined(_MSC_VER)
  delete FindFileData;
#endif
}

bool VListFileDir::Open(const char * Path)
{
  if (strlen(Path) + 10 >= sizeof(FullPath))
    return false;

  strcpy(FullPath, Path);
  if (FullPath[0] && FullPath[strlen(FullPath)-1] != '/')
    strcat(FullPath, "/");
  PathLen = strlen(FullPath);
#if defined(__VC__) || defined(_MSC_VER)
  strcat(FullPath, "*.*");
  dir = FindFirstFileA(FullPath, (WIN32_FIND_DATA *)FindFileData);
  FirstRead = true;
  FullPath[PathLen] = 0;
#else
  dir = opendir(Path);
#endif

  return dir != NULL;
}

bool VListFileDir::Open(const char * Path, const char * RelPath)
{
#if defined(__WINDOWS__)
  char Dir[2048];
#else
  char Dir[strlen(Path) + strlen(RelPath) + 2];
#endif

  strcpy(Dir, Path);
  strcat(Dir, RelPath);

  return Open(Dir);
}

void VListFileDir::Close()
{
#if defined(__VC__) || defined(_MSC_VER)
  if (dir)
    FindClose((HANDLE)dir);
#else
  if (dir)
    closedir((DIR *)dir);
#endif
  dir = NULL;
}

bool VListFileDir::Next()
{
  if (!dir)
    return false;

#if defined(__VC__) || defined(_MSC_VER)
  if (!FirstRead)
  {
    if (!FindNextFile((HANDLE) dir, (WIN32_FIND_DATA *)FindFileData))
      return false;
  }
  FirstRead = false;
  strcpy(Name, ((WIN32_FIND_DATA *)FindFileData)->cFileName);
#else
  dirent * dire = readdir((DIR *)dir);
  if (!dire)
    return false;

  strcpy(Name, dire->d_name);
#endif
  if (PathLen + strlen(Name) + 1 >= sizeof(FullPath))
    return false;
  strcpy(FullPath + PathLen, Name);

  mode = size = time = 0;
  struct stat Stat;
  if (!stat(FullPath, &Stat))
  {
    mode = Stat.st_mode;
    size = Stat.st_size;
    time = Stat.st_mtime;
  }

  return true;
}

bool VListFileDir::IsFile() const
{
#if defined(__VC__) || defined(_MSC_VER)
  if (((WIN32_FIND_DATA *)FindFileData)->dwFileAttributes &
      (FILE_ATTRIBUTE_SYSTEM | FILE_ATTRIBUTE_TEMPORARY))
    return false;
#endif
#if defined(__ARM__)
  if (!mode)
    return true;
#endif
  return !(mode & S_IFDIR) &&
         (mode & S_IFREG);
}

bool VListFileDir::IsDir() const
{
#if defined(__VC__) || defined(_MSC_VER)
  if (((WIN32_FIND_DATA *)FindFileData)->dwFileAttributes &
      (FILE_ATTRIBUTE_SYSTEM | FILE_ATTRIBUTE_TEMPORARY))
    return false;
#endif
  return (mode & S_IFDIR) != 0;
}
